// Bibliotheken zum Einbinden (Zufallszahlen, Ein-Ausgabe)
#include <cstdlib>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include "hilf.h"
#include "vector.h"
#include "graphen.h"
#include "kuerzwege.h"
#include "network.h"
using namespace std;

void findweight(int* column_indices,double* entries,
              int* row_offsets,int i,int j,double *wij)
// Bestimmt das Gewicht einer Kante (i,j)
// Input: einfacher (Rest)graph im CSR-Format, Kante (i,j)
// Output: Gewicht wij (ist Null falls keine Kante vorhanden)
{
  int j1,k,k1,k2;

  *wij=0;
  k1=row_offsets[i];
  k2=row_offsets[i+1];
  k=k1;
  // Durchsuche alle Nachbarn von i nach j
  while (k<k2)
    {
    j1=column_indices[k];
    if (j1==j)
      {
        *wij=entries[k];
        k=k2;  //j Gefunden, Abbruch der Schleife
      }
    k++;
     }
}

void mincap(int* column_indices,double* entries,
          int* row_offsets,int* p,double *gamma,int s,int t)
// Finde minimales Gewicht entlang des s-t Weges p
// Input: einfacher RestGraph in CSR-Format,s, t, 
//    Weg p als  Vorgaenger gespeichert
// (muss existieren)
// Output:gamma
{
  int i,j;
  const double oxer=pow(10,40);
  double wij;
  *gamma=oxer;
  j=t;
  i=p[j];
  findweight(column_indices,entries,row_offsets,i,j,&wij);
  if (wij<*gamma)       
    {
     *gamma=wij;
    }
  while (i!=s)
    { 
    j=i;
    i=p[j]; 
    findweight(column_indices,entries,row_offsets,i,j,&wij);
     if (wij<*gamma)       
       {
	 *gamma=wij;
       }
    }
}

void Computeflow(int* column_indices,double* flow,
       int* row_offsets,double* capacity,
       int m,int n,double* nodeflow)
// Berechnet den Fluss in allen Knoten: Nodeflow
// Input: flow: Fluss Im letzten Restgraphen
//        capacity: Kapazitaeten 
// Jeweils als CSR-Format gespeichert mit Hin und Rueckkante 
//                     evtl. 0
// d.h. flow->capacity-flow ist aktueller Fluss (falls c>0)
// und 0 auf allen Rueckwaertskanten
{
  double* e;
  e=(double*)calloc(n,sizeof(double));
  vdcopy(n,e,1); // e=(1,1,1,...,1)
  vd0mult(m,flow,capacity,flow); 
        // Entferne alle Rueckwaertskanten
  vdminus(m,capacity,flow,flow); 
  MatVecMulCSR(column_indices,flow,row_offsets,e,nodeflow,n,m);
  // Berechnung als nodeflow_i=\sum_{(i,j)\in E} flow(i,j) 
  // D.h. durch Multiplikation der Matrix flow(i,j) mit e 
}

void EdmondsKarp(int* column_indices,double* entries,
         int* row_offsets,int s,int t,int n)
// funktioniert nicht bei Restgraphen mit parallelen Kanten.
// Input ger. Graph, Rueckwaertskante hat Gewicht 0.
// Kapazitaeten werden uberschrieben
{
  int m,k;
  int* p;
  double gamma,flow;
  
  p=(int*)calloc(n,sizeof(int));
  m=row_offsets[n]; //anzahl an Kanten
  BFS(n,s,column_indices,entries,row_offsets,p);
  k=0; // Iterationszaehler
  flow=0; // Fluss
  while (p[t]>=0)  // p[t]=-1, d.h. t ist nicht erreichbar
    {
      mincap(column_indices,entries,row_offsets,p,&gamma,s,t);
      flow=flow+gamma; 
      cout <<"iteration "<< k <<" Fluss "<<flow <<endl;
      k++;
      aug(column_indices,entries,row_offsets,p,gamma,s,t);
      BFS(n,s,column_indices,entries,row_offsets,p);
      if (n<10)             // Ausgabe augmentierender Weg
	{
	  ausgabeint(p,n);
        }
    }
  free(p);
}


// Das Hauptprogramm
// Fluss in den Knoten wird (bei Gitter) in Datei flow.dat geschrieben
int main()
{ const int nmax=100000;
  const int nnz=1000000; 
 
  int* column_indices;
  int* row_offsets;
  int i,j,k,ch,n,s,t,m,nz,m2,m1;
  double* entries;
  double* fluss;
  double* nodeflow;
  // Einlesen des Graphen
  cout <<"Art des Graphen: Gitter (1), vollst. (2)" << endl;
  cin >> ch;
  if (ch==1)
    { 
     cout <<"Groesse des Gitters (axb)"<< endl;
     cin >> m1>> m2; 
     n=m1*m2;
     nz=4*n; 
    }
  else
    {
     cout <<"Groesse des Graphen"<< endl;
     cin >> m1; 
     n=m1;
     nz=(n-1)*n;
    }
   column_indices=(int*)malloc(nz*sizeof(int));   
   row_offsets=(int*)malloc((n+1)*sizeof(int));   
   entries=(double*)malloc(nz*sizeof(double));       
   fluss=(double*)malloc(nz*sizeof(double));   
   nodeflow=(double*)malloc(n*sizeof(double));
   if (ch==1)
     {
      Netzwerk_gitter(m1,m2,column_indices,entries,row_offsets);
     }
   else
      {
      zufall(entries,nz);
      Netzwerk_vollst(m1,column_indices,entries,row_offsets);      
     }
   m=row_offsets[n];
   s=0;
   t=n-1;
   vdcopyvector(m,fluss,entries);
   EdmondsKarp(column_indices,fluss,row_offsets,s,t,n);
   // Ergebnisausgabe
   // Aufbereitung fuer Graphik, 
   //  Bestimmung Fluss durch die Knoten
   Computeflow(column_indices,fluss,row_offsets,entries,
              m,n,nodeflow);
   if (n<300)  
     {
            ausgabe(entries,m);
          ausgabe(fluss,m);
       cout << " Fluss in den Knoten" << endl;
     ausgabe(nodeflow,n); 
     }
   if (ch==1)
     {
   fstream f;
   f.open("flow.dat", fstream::out);
   for (i=0;i<m1;i++)
     for (j=0;j<m2;j++)
     {
       f << i << " \t" << j <<" \t" <<nodeflow[k++] << endl;
     }   
   f.close();
     }
   free(entries);
   free(column_indices);
   free(row_offsets);
  return 0;
}
